"""Configuration.from_pydantic() tests."""

import pydantic
from dependency_injector import providers, errors
from pytest import fixture, mark, raises


class Section11(pydantic.BaseModel):
    value1 = 1


class Section12(pydantic.BaseModel):
    value2 = 2


class Settings1(pydantic.BaseSettings):
    section1 = Section11()
    section2 = Section12()


class Section21(pydantic.BaseModel):
    value1 = 11
    value11 = 11


class Section3(pydantic.BaseModel):
    value3 = 3


class Settings2(pydantic.BaseSettings):
    section1 = Section21()
    section3 = Section3()

@fixture
def no_pydantic_module_installed():
    providers.pydantic = None
    yield
    providers.pydantic = pydantic


def test(config):
    config.from_pydantic(Settings1())

    assert config() == {"section1": {"value1": 1}, "section2": {"value2": 2}}
    assert config.section1() == {"value1": 1}
    assert config.section1.value1() == 1
    assert config.section2() == {"value2": 2}
    assert config.section2.value2() == 2


def test_kwarg(config):
    config.from_pydantic(Settings1(), exclude={"section2"})

    assert config() == {"section1": {"value1": 1}}
    assert config.section1() == {"value1": 1}
    assert config.section1.value1() == 1


def test_merge(config):
    config.from_pydantic(Settings1())
    config.from_pydantic(Settings2())

    assert config() == {
        "section1": {
            "value1": 11,
            "value11": 11,
        },
        "section2": {
            "value2": 2,
        },
        "section3": {
            "value3": 3,
        },
    }
    assert config.section1() == {"value1": 11, "value11": 11}
    assert config.section1.value1() == 11
    assert config.section1.value11() == 11
    assert config.section2() == {"value2": 2}
    assert config.section2.value2() == 2
    assert config.section3() == {"value3": 3}
    assert config.section3.value3() == 3


def test_empty_settings(config):
    config.from_pydantic(pydantic.BaseSettings())
    assert config() == {}


@mark.parametrize("config_type", ["strict"])
def test_empty_settings_strict_mode(config):
    with raises(ValueError):
        config.from_pydantic(pydantic.BaseSettings())


def test_option_empty_settings(config):
    config.option.from_pydantic(pydantic.BaseSettings())
    assert config.option() == {}


@mark.parametrize("config_type", ["strict"])
def test_option_empty_settings_strict_mode(config):
    with raises(ValueError):
        config.option.from_pydantic(pydantic.BaseSettings())


def test_required_empty_settings(config):
    with raises(ValueError):
        config.from_pydantic(pydantic.BaseSettings(), required=True)


def test_required_option_empty_settings(config):
    with raises(ValueError):
        config.option.from_pydantic(pydantic.BaseSettings(), required=True)


@mark.parametrize("config_type", ["strict"])
def test_not_required_empty_settings_strict_mode(config):
    config.from_pydantic(pydantic.BaseSettings(), required=False)
    assert config() == {}


@mark.parametrize("config_type", ["strict"])
def test_not_required_option_empty_settings_strict_mode(config):
    config.option.from_pydantic(pydantic.BaseSettings(), required=False)
    assert config.option() == {}
    assert config() == {"option": {}}


def test_not_instance_of_settings(config):
    with raises(errors.Error) as error:
        config.from_pydantic({})
    assert error.value.args[0] == (
        "Unable to recognize settings instance, expect \"pydantic.BaseSettings\", "
        "got {0} instead".format({})
    )


def test_option_not_instance_of_settings(config):
    with raises(errors.Error) as error:
        config.option.from_pydantic({})
    assert error.value.args[0] == (
        "Unable to recognize settings instance, expect \"pydantic.BaseSettings\", "
        "got {0} instead".format({})
    )


def test_subclass_instead_of_instance(config):
    with raises(errors.Error) as error:
        config.from_pydantic(Settings1)
    assert error.value.args[0] == (
        "Got settings class, but expect instance: "
        "instead \"Settings1\" use \"Settings1()\""
    )


def test_option_subclass_instead_of_instance(config):
    with raises(errors.Error) as error:
        config.option.from_pydantic(Settings1)
    assert error.value.args[0] == (
        "Got settings class, but expect instance: "
        "instead \"Settings1\" use \"Settings1()\""
    )


@mark.usefixtures("no_pydantic_module_installed")
def test_no_pydantic_installed(config):
    with raises(errors.Error) as error:
        config.from_pydantic(Settings1())
    assert error.value.args[0] == (
        "Unable to load pydantic configuration - pydantic is not installed. "
        "Install pydantic or install Dependency Injector with pydantic extras: "
        "\"pip install dependency-injector[pydantic]\""
    )


@mark.usefixtures("no_pydantic_module_installed")
def test_option_no_pydantic_installed(config):
    with raises(errors.Error) as error:
        config.option.from_pydantic(Settings1())
    assert error.value.args[0] == (
        "Unable to load pydantic configuration - pydantic is not installed. "
        "Install pydantic or install Dependency Injector with pydantic extras: "
        "\"pip install dependency-injector[pydantic]\""
    )



from pathlib import Path
from dependency_injector import providers


# Test 1: to check the getattr
config1 = providers.Configuration()
provider1 = providers.Factory(dict, a=config1.a)

# Test 2: to check the from_*() method
config2 = providers.Configuration()
config2.from_dict({})
config2.from_ini("config.ini")
config2.from_ini(Path("config.ini"))

config2.from_yaml("config.yml")
config2.from_yaml(Path("config.yml"))

config2.from_env("ENV", "default")
config2.from_env("ENV", as_=int, default=123)
config2.from_env("ENV", as_=float, required=True)
config2.from_env("ENV", as_=lambda env: str(env))

# Test 3: to check as_*() methods
config3 = providers.Configuration()
int3: providers.Callable[int] = config3.option.as_int()
float3: providers.Callable[float] = config3.option.as_float()
int3_custom: providers.Callable[int] = config3.option.as_(int)

# Test 4: to check required() method
config4 = providers.Configuration()
option4: providers.ConfigurationOption = config4.option.required()



"""Wiring optimizations module."""

import asyncio
import collections.abc
import functools
import inspect
import types

from . import providers
from .wiring import _Marker


def _get_sync_patched(fn):
    @functools.wraps(fn)
    def _patched(*args, **kwargs):
        cdef object result
        cdef dict to_inject

        to_inject = kwargs.copy()
        for injection, provider in _patched.__injections__.items():
            if injection not in kwargs or isinstance(kwargs[injection], _Marker):
                to_inject[injection] = provider()

        result = fn(*args, **to_inject)

        if _patched.__closing__:
            for injection, provider in _patched.__closing__.items():
                if injection in kwargs and not isinstance(kwargs[injection], _Marker):
                    continue
                if not isinstance(provider, providers.Resource):
                    continue
                provider.shutdown()

        return result
    return _patched


def _get_async_patched(fn):
    @functools.wraps(fn)
    async def _patched(*args, **kwargs):
        cdef object result
        cdef dict to_inject
        cdef list to_inject_await = []
        cdef list to_close_await = []

        to_inject = kwargs.copy()
        for injection, provider in _patched.__injections__.items():
            if injection not in kwargs or isinstance(kwargs[injection], _Marker):
                provide = provider()
                if _isawaitable(provide):
                    to_inject_await.append((injection, provide))
                else:
                    to_inject[injection] = provide

        if to_inject_await:
            async_to_inject = await asyncio.gather(*(provide for _, provide in to_inject_await))
            for provide, (injection, _) in zip(async_to_inject, to_inject_await):
                to_inject[injection] = provide

        result = await fn(*args, **to_inject)

        if _patched.__closing__:
            for injection, provider in _patched.__closing__.items():
                if injection in kwargs \
                        and isinstance(kwargs[injection], _Marker):
                    continue
                if not isinstance(provider, providers.Resource):
                    continue
                shutdown = provider.shutdown()
                if _isawaitable(shutdown):
                    to_close_await.append(shutdown)

            await asyncio.gather(*to_close_await)

        return result

    # Hotfix for iscoroutinefunction() for Cython < 3.0.0; can be removed after migration to Cython 3.0.0+
    _patched._is_coroutine = asyncio.coroutines._is_coroutine

    return _patched


cdef bint _isawaitable(object instance):
    """Return true if object can be passed to an ``await`` expression."""
    return (isinstance(instance, types.CoroutineType) or
            isinstance(instance, types.GeneratorType) and
            bool(instance.gi_code.co_flags & inspect.CO_ITERABLE_COROUTINE) or
            isinstance(instance, collections.abc.Awaitable))



"""`Dependency injector` setup script."""

import os
import re

from setuptools import setup, Extension


# Defining setup variables:
defined_macros = dict()
defined_macros["CYTHON_CLINE_IN_TRACEBACK"] = 0

# Getting description:
with open("README.rst") as readme_file:
    description = readme_file.read()

# Getting requirements:
with open("requirements.txt") as requirements_file:
    requirements = requirements_file.readlines()

# Getting version:
with open("src/dependency_injector/__init__.py") as init_file:
    version = re.search("__version__ = \"(.*?)\"", init_file.read()).group(1)

# Adding debug options:
if os.environ.get("DEPENDENCY_INJECTOR_DEBUG_MODE") == "1":
    defined_macros["CYTHON_TRACE"] = 1
    defined_macros["CYTHON_TRACE_NOGIL"] = 1
    defined_macros["CYTHON_CLINE_IN_TRACEBACK"] = 1


setup(name="dependency-injector",
      version=version,
      description="Dependency injection framework for Python",
      long_description=description,
      author="Roman Mogylatov",
      author_email="rmogilatov@gmail.com",
      maintainer="Roman Mogylatov",
      maintainer_email="rmogilatov@gmail.com",
      url="https://github.com/ets-labs/python-dependency-injector",
      download_url="https://pypi.python.org/pypi/dependency_injector",
      packages=[
          "dependency_injector",
          "dependency_injector.ext",
      ],
      package_dir={
          "": "src",
      },
      package_data={
          "dependency_injector": ["*.pxd", "*.pyi", "py.typed"],
      },
      ext_modules=[
          Extension("dependency_injector.containers",
                    ["src/dependency_injector/containers.c"],
                    define_macros=list(defined_macros.items()),
                    extra_compile_args=["-O2"]),
          Extension("dependency_injector.providers",
                    ["src/dependency_injector/providers.c"],
                    define_macros=list(defined_macros.items()),
                    extra_compile_args=["-O2"]),
          Extension("dependency_injector._cwiring",
                    ["src/dependency_injector/_cwiring.c"],
                    define_macros=list(defined_macros.items()),
                    extra_compile_args=["-O2"]),
      ],
      install_requires=requirements,
      extras_require={
          "yaml": [
              "pyyaml",
          ],
          "pydantic": [
              "pydantic",
          ],
          "flask": [
              "flask",
          ],
          "aiohttp": [
              "aiohttp",
          ],
      },
      zip_safe=True,
      license="BSD New",
      platforms=["any"],
      keywords=[
          "Dependency injection",
          "DI",
          "Inversion of Control",
          "IoC",
          "Factory",
          "Singleton",
          "Design patterns",
          "Flask",
      ],
      classifiers=[
          "Development Status :: 5 - Production/Stable",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: BSD License",
          "Operating System :: OS Independent",
          "Programming Language :: Python",
          "Programming Language :: Python :: 2",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.5",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "Programming Language :: Python :: 3.8",
          "Programming Language :: Python :: 3.9",
          "Programming Language :: Python :: 3.10",
          "Programming Language :: Python :: Implementation :: CPython",
          "Programming Language :: Python :: Implementation :: PyPy",
          "Framework :: AsyncIO",
          "Framework :: Bottle",
          "Framework :: Django",
          "Framework :: Flask",
          "Framework :: Pylons",
          "Framework :: Pyramid",
          "Framework :: Pytest",
          "Framework :: TurboGears",
          "Topic :: Software Development",
          "Topic :: Software Development :: Libraries",
          "Topic :: Software Development :: Libraries :: Python Modules",
      ])
